﻿IncludeFile "Token.pb"
IncludeFile "InterpreterError.pb"

;
; File: Lexer.pb
; Copyright (C) Philipp Isserstedt, 2011
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
; 
; This software is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY Or FITNESS for A PARTICULAR PURPOSE. See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License.
; If not, see <http://www.gnu.org/licenses/>.
;

; -----------------------------------------------------
;-> Constants
; -----------------------------------------------------

#CharSize = SizeOf(Character)

; -----------------------------------------------------
;-> Macros & Procedures
; -----------------------------------------------------

;
; Macro to save typing while assign values to an token object.
;
Macro SetValues(_Token, _Type, _Name, _Value = NaN())
	_Token\Type = _Type
	_Token\Name = _Name
	_Token\Value = _Value
EndMacro

;
; Checks whether the given character is a whitespace character, i.e. TAB, CR, LF, Space.
; Returns #True If the given character is a whitespace character, otherwise #False.
;
; Parameters:
;		Char : The character which should be tested.
;
Procedure IsSpace(Char.c)
	If Char = 9 Or Char = 10 Or Char = 13 Or Char = 32
		ProcedureReturn #True
	EndIf
	ProcedureReturn #False
EndProcedure

;
; Checks whether the given character is a digit, i.e. '0',...,'9'.
; Returns #True if the given character is a digit, otherwise #False.
;
; Parameters:
;		Char : The character which should be tested.
;
Procedure IsDigit(Char.c)
	If Char >= '0' And Char <= '9'
		ProcedureReturn #True
	EndIf
	ProcedureReturn #False
EndProcedure

;
; Checks whether the given character is a letter, i.e. 'a',...,'z','A',...,'Z'.
; Returns #True if the given character is a letter, otherwise #False.
;
; Parameters:
;		Char : The character which should be tested.
;
Procedure IsLetter(Char.c)
	If (Char >= 'a' And Char <= 'z') Or (Char >= 'A' And Char <= 'Z')
		ProcedureReturn #True
	EndIf
	ProcedureReturn #False
EndProcedure

;
; Gets the next token from the formula string. Returns the current pointer position
; within the input string if no error occures, otherwise 0.
;
; Parameters:
;		*Input : Character pointer to the formula-string.
;		*Token : A pointer to a Token-structure to store the information of the next token.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
Procedure GetNextToken(*Input.Character, *Token.Token, *Error.InterpreterError)
	Protected Buffer.s = ""
	While IsSpace(*Input\c)
		*Input + #CharSize
	Wend
	Select *Input\c
		Case 0
			SetValues(*Token, #TokenType_EOF, TokenTypeString(#TokenType_EOF))
		Case '+'
			SetValues(*Token, #TokenType_Plus, TokenTypeString(#TokenType_Plus))
			*Input + #CharSize
		Case '-'
			SetValues(*Token, #TokenType_Minus, TokenTypeString(#TokenType_Minus))
			*Input + #CharSize
		Case '*'
			SetValues(*Token, #TokenType_Dot, TokenTypeString(#TokenType_Dot))
			*Input + #CharSize
		Case '/'
			SetValues(*Token, #TokenType_Slash, TokenTypeString(#TokenType_Slash))
			*Input + #CharSize
		Case ','
			SetValues(*Token, #TokenType_Comma, TokenTypeString(#TokenType_Comma))
			*Input + #CharSize
		Case '('
			SetValues(*Token, #TokenType_Bra, TokenTypeString(#TokenType_Bra))
			*Input + #CharSize
		Case ')'
			SetValues(*Token, #TokenType_Ket, TokenTypeString(#TokenType_Ket))
			*Input + #CharSize
		Case 'a' To 'z', 'A' To 'Z', '_'
			While IsLetter(*Input\c) Or *Input\c = '_'
				Buffer + Chr(*Input\c)
				*Input + #CharSize
			Wend
			SetValues(*Token, #TokenType_Identifier, Buffer)
		Case '0' To '9'
			While IsDigit(*Input\c)
				Buffer + Chr(*Input\c)
				*Input + #CharSize
			Wend
			If *Input\c = '.'
				Buffer + Chr(*Input\c)
				*Input + #CharSize
				If Not IsDigit(*Input\c)
					*Error\Type = #InterpreterError_InvalidNumberFormat
					*Error\Info1 = Buffer + Chr(*Input\c)
					ProcedureReturn 0
				EndIf
				While IsDigit(*Input\c)
					Buffer + Chr(*Input\c)
					*Input + #CharSize
				Wend
			EndIf
			SetValues(*Token, #TokenType_Number, Buffer, ValD(Buffer))
		Default
			*Error\Type = #InterpreterError_InvalidSign
			*Error\Info1 = Chr(*Input\c)
			ProcedureReturn 0
	EndSelect
	ProcedureReturn *Input
EndProcedure

;
; Splits the given formula into tokens and stores the information in the given list.
; Returns #True if no error occures, otherwise #False.
;
; Parameters:
;		Formula : The formula-string which should be calculated.
;		Tokens : A list in which the tokens are stored.
;		*Error : Pointer to a InterpreterError-structure to store possible error information.
;
Procedure Tokenize(Formula.s, List Tokens.Token(), *Error.InterpreterError)
	Protected *Input.Character
	Protected Token.Token
	ClearList(Tokens())
	*Input = GetNextToken(@Formula, @Token, *Error)
	If Not *Input
		ProcedureReturn #False
	EndIf
	AddElement(Tokens())
	SetValues(Tokens(), Token\Type, Token\Name, Token\Value)
	While Token\Type <> #TokenType_EOF
		*Input = GetNextToken(*Input, @Token, *Error)
		If Not *Input
			ProcedureReturn #False
		EndIf
		AddElement(Tokens())
		SetValues(Tokens(), Token\Type, Token\Name, Token\Value)
	Wend
	FirstElement(Tokens())
	ProcedureReturn #True
EndProcedure

; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 19
; FirstLine = 3
; Folding = --
; EnableUnicode
; EnableXP
; CompileSourceDirectory