﻿;
; File: Parser.pb
; Copyright (C) Philipp Isserstedt, 2011
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
; 
; This software is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY Or FITNESS for A PARTICULAR PURPOSE. See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License.
; If not, see <http://www.gnu.org/licenses/>.
;

IncludeFile "Lexer.pb"
IncludeFile "InterpreterEnvironment.pb"

; -----------------------------------------------------
;-> Procedures
; -----------------------------------------------------

Declare.d Multiplication(List Tokens.Token(), *Environment.InterpreterEnvironment, *Error.InterpreterError)
Declare.d Parenthesis(List Tokens.Token(), *Environment.InterpreterEnvironment, *Error.InterpreterError)
Declare.d Function(List Tokens.Token(), *Environment.InterpreterEnvironment, *Error.InterpreterError)
Declare.d Identifier(List Tokens.Token(), *Environment.InterpreterEnvironment, *Error.InterpreterError)
Declare.d Number(List Tokens.Token(), *Error.InterpreterError)

;
; Evaluates a identifier and returns the result. In case of an error, NaN() is returned.
;
; Parameters:
;		Name : The name of the identifier.
;		Constants() : All predefined constants.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
Procedure.d EvalIdentifier(Name.s, Map Constants.d(), *Error.InterpreterError)
	If Not FindMapElement(Constants(), Name)
		*Error\Type = #InterpreterError_UnknownIdentifier
		*Error\Info1 = Name
		ProcedureReturn NaN()
	EndIf
	ProcedureReturn Constants()
EndProcedure

;
; Calculates a function call and returns the result. In case of an error, NaN() is returned.
;
; Parameters:
;		Name : The name of the function.
;		Arguments() : The arguments with which the function was called.
;		Functions() : All predefined functions.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
Procedure.d EvalFunction(Name.s, List Arguments.d(), Map Functions.Function(), *Error.InterpreterError)
	If Not FindMapElement(Functions(), Name)
		*Error\Type = #InterpreterError_UnknownFunction
		*Error\Info1 = Name
		ProcedureReturn NaN()
	EndIf
	If Functions()\NumberOfArguments <> ListSize(Arguments())
		*Error\Type = #InterpreterError_InvalidNumberOfArguments
		*Error\Info1 = Name
		*Error\Info2 = Str(ListSize(Arguments())) + "|" + Str(Functions()\NumberOfArguments)		; Format: found|expected
		ProcedureReturn NaN()
	EndIf
	Dim ArgumentArray.d(ListSize(Arguments()) - 1)
	Protected Index = 0
	ForEach Arguments()
		ArgumentArray(Index) = Arguments()
		Index + 1
	Next
	ProcedureReturn Functions()\Eval(ArgumentArray())
EndProcedure

;
; Returns #True if the expected token corresponds to the current, otherwise #False.
; In addition, the next token is retrieved from the token-list.
;
; Parameters:
;		ExpectedType : The expected token type.
;		Tokens() : The tokenized formula-string.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
Procedure Expect(ExpectedType.b, List Tokens.Token(), *Error.InterpreterError)
	Protected Result = #True
	If ExpectedType <> Tokens()\Type
		*Error\Type = #InterpreterError_UnexpectedToken
		*Error\Info1 = TokenTypeString(Tokens()\Type)		; found
		*Error\Info2 = TokenTypeString(ExpectedType)		; expected
		Result = #False
	EndIf
	NextElement(Tokens())
	ProcedureReturn Result
EndProcedure

;
; Returns a evaluated block-expression with respect to the following EBNF:
; block = (multiplication) { ("+" | "-") (multiplication) };
; In case of an error, NaN() is returned.
;
; Parameters:
;		Tokens() : The tokenized formula-string.
;		*Environment : A pointer to a Environment-structure to provide information about constants and functions.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
Procedure.d Block(List Tokens.Token(), *Environment.InterpreterEnvironment, *Error.InterpreterError)
	Protected Result.d = Multiplication(Tokens(), *Environment, *Error)
	While Tokens()\Type = #TokenType_Plus Or Tokens()\Type = #TokenType_Minus
		Select Tokens()\Type
			Case #TokenType_Plus
				NextElement(Tokens())
				Result + Multiplication(Tokens(), *Environment, *Error)
			Case #TokenType_Minus
				NextElement(Tokens())
				Result - Multiplication(Tokens(), *Environment, *Error)
		EndSelect
	Wend
	ProcedureReturn Result
EndProcedure

;
; Returns a evaluated multiplication-expression with respect to the following EBNF:
; multiplication = (parenthesis) { ("*" | "/") (parenthesis) };
; In case of an error, NaN() is returned.
;
; Parameters:
;		Tokens() : The tokenized formula-string.
;		*Environment : A pointer to a Environment-structure to provide information about constants and functions.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
Procedure.d Multiplication(List Tokens.Token(), *Environment.InterpreterEnvironment, *Error.InterpreterError)
	Protected Divisor.d = 0
	Protected Result.d = Parenthesis(Tokens(), *Environment, *Error)
	While Tokens()\Type = #TokenType_Dot Or Tokens()\Type = #TokenType_Slash
		Select Tokens()\Type
			Case #TokenType_Dot
				NextElement(Tokens())
				Result * Parenthesis(Tokens(), *Environment, *Error)
			Case #TokenType_Slash
				NextElement(Tokens())
				Divisor.d = Parenthesis(Tokens(), *Environment, *Error)
				If Divisor = 0
					*Error\Type = #InterpreterError_DivisionByZero
					ProcedureReturn NaN()
				Else
					Result / Divisor
				EndIf
		EndSelect
	Wend
	ProcedureReturn Result
EndProcedure

;
; Returns a evaluated parenthesis-expression with respect to the following EBNF:
; parenthesis = ["+" | "-"] ( (function) | (identifier) | (number) | "(" (block) ")" );
; In case of an error, NaN() is returned.
;
; Parameters:
;		Tokens() : The tokenized formula-string.
;		*Environment : A pointer to a Environment-structure to provide information about constants and functions.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
Procedure.d Parenthesis(List Tokens.Token(), *Environment.InterpreterEnvironment, *Error.InterpreterError)
	Protected Result.d = NaN()
	Protected Sign = 1
	Protected *Lookup.Token = 0
	If Tokens()\Type = #TokenType_Plus Or Tokens()\Type = #TokenType_Minus
		If Tokens()\Type = #TokenType_Minus
			Sign = -1
		EndIf
		NextElement(Tokens())
	EndIf
	If Tokens()\Type = #TokenType_Bra
		If Not Expect(#TokenType_Bra, Tokens(), *Error)
			ProcedureReturn NaN()
		EndIf
		Result = Block(Tokens(), *Environment, *Error)
		If Not Expect(#TokenType_Ket, Tokens(), *Error)
			ProcedureReturn NaN()
		EndIf
		ProcedureReturn Sign * Result
	ElseIf Tokens()\Type = #TokenType_Identifier
		*Lookup.Token = NextElement(Tokens())
		PreviousElement(Tokens())
		If *Lookup And *Lookup\Type = #TokenType_Bra
			ProcedureReturn Sign * Function(Tokens(), *Environment, *Error)
		EndIf
		ProcedureReturn Sign * Identifier(Tokens(), *Environment, *Error)
	EndIf
	ProcedureReturn Sign * Number(Tokens(), *Error)
EndProcedure

;
; Returns a evaluated function from the current token-list with respect
; to the following EBNF: function = (identifier) "(" [ (block) { "," (block) } ] ")";
; In case of an error, NaN() is returned.
;
; Parameters:
;		Tokens() : The tokenized formula-string.
;		*Environment : A pointer to a Environment-structure to provide information about constants and functions.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
Procedure.d Function(List Tokens.Token(), *Environment.InterpreterEnvironment, *Error.InterpreterError)
	Protected Name.s = Tokens()\Name
	NewList Arguments.d()
	If Not Expect(#TokenType_Identifier, Tokens(), *Error) Or Not Expect(#TokenType_Bra, Tokens(), *Error)
		ProcedureReturn NaN()
	EndIf
	AddElement(Arguments())
	Arguments() = Block(Tokens(), *Environment, *Error)
	While Tokens()\Type = #TokenType_Comma
		If Not Expect(#TokenType_Comma, Tokens(), *Error)
			ProcedureReturn NaN()
		EndIf
		AddElement(Arguments())
		Arguments() = Block(Tokens(), *Environment, *Error)
	Wend
	If Not Expect(#TokenType_Ket, Tokens(), *Error)
		ProcedureReturn NaN()
	EndIf
	ProcedureReturn EvalFunction(Name, Arguments(), *Environment\Functions(), *Error)
EndProcedure

;
; Returns a evaluated identifier (i.e. a constant) from the current token-list with
; respect to the following EBNF: identifier = ( ("a"-"z") | "_" { ("a"-"z") | "_" } );
; In case of an error, NaN() is returned.
;
; Parameters:
;		Tokens() : The tokenized formula-string.
;		*Environment : A pointer to a Environment-structure to provide information about constants and functions.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
Procedure.d Identifier(List Tokens.Token(), *Environment.InterpreterEnvironment, *Error.InterpreterError)
	Protected Result.d = EvalIdentifier(Tokens()\Name, *Environment\Constants(), *Error)
	If Not Expect(#TokenType_Identifier, Tokens(), *Error)
		Result = NaN()
	EndIf
	ProcedureReturn Result
EndProcedure

;
; Returns a number from the current token-list with respect
; to the following EBNF: number = (digit) { (digit) };
;												 digit = "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9";
; In case of an error, NaN() is returned.
;
; Parameters:
;		Tokens() : The tokenized formula-string.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
Procedure.d Number(List Tokens.Token(), *Error.InterpreterError)
	Protected Result.d = Tokens()\Value
	If Not Expect(#TokenType_Number, Tokens(), *Error)
		Result = NaN()
	EndIf
	ProcedureReturn Result
EndProcedure

;
; Calculates the given formula. Returns the result of the formula if no error occurred, otherwise NaN().
;
; Parameters:
;		Formula : The mathematical expression which should be calculated.
;		*Environment : A pointer to a Environment-structure to provide information about constants and functions.
;		*Error : A pointer to a InterpreterError-structure to store possible error information.
;
ProcedureDLL.d Parse(Formula.s, *Environment.InterpreterEnvironment, *Error.InterpreterError)
	Protected Result.d = NaN()
	NewList Tokens.Token()
	If Not Tokenize(Formula, Tokens(), *Error)
		ProcedureReturn Result
	EndIf
	Result = Block(Tokens(), *Environment, *Error)
	If Not Expect(#TokenType_EOF, Tokens(), *Error)
		Result = NaN()
	EndIf
	ProcedureReturn Result
EndProcedure

; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 195
; FirstLine = 238
; Folding = ---
; EnableUnicode
; EnableXP
; CompileSourceDirectory